
#include <nds.h>

#include <stdio.h>
#include <stdlib.h>

#include "_console.h"
#include "_consolewritelog.h"
#include "maindef.h"
#include "memtool.h"
#include "_const.h"
#include "inifile.h"

#include "glib.h"
#include "rect.h"
#include "strtool.h"

#include "fat2.h"
#include "shell.h"
#include "procstate.h"

#include "skinpack.h"

#include "Particle.h"

static bool Enabled;

enum EItemType {EIT_Cursor,EIT_Fire,EIT_Bomb};

typedef struct {
  u32 LifeTime;
  EItemType EIT;
  s32 fx,fy;
  s32 fvx,fvy;
} TItem;
static const u32 ItemsCount=256;
static TItem Items[ItemsCount];

void Particle_Init(bool _Enabled)
{
  Enabled=_Enabled;
  if(ProcState.Main.UseScreenEffect==false) Enabled=false;
  
  for(u32 idx=0;idx<ItemsCount;idx++){
    TItem *pi=&Items[idx];
    pi->LifeTime=0;
  }
}

void Particle_Free(void)
{
}

void Particle_Process(void)
{
  for(u32 idx=0;idx<ItemsCount;idx++){
    TItem *pi=&Items[idx];
    if(pi->LifeTime!=0){
      pi->LifeTime--;
      pi->fvx=(pi->fvx*63)/64;
      pi->fvy+=0.02*0x100;
      pi->fx+=pi->fvx;
      pi->fy+=pi->fvy;
      const u32 fw=(ScreenWidth-1)*0x100,fh=(ScreenHeight-1)*0x100;
      if((pi->fx<(1*0x100))||(fw<pi->fx)) pi->LifeTime=0;
      if((pi->fy<(1*0x100))||(fh<pi->fy)) pi->LifeTime=0;
    }
  }
}

void Particle_Draw(CglCanvas *pcan)
{
  u16 *pcanbuf=pcan->GetVRAMBuf();
  
  for(u32 idx=0;idx<ItemsCount;idx++){
    TItem *pi=&Items[idx];
    const u32 LT=pi->LifeTime;
    if(LT!=0){
      switch(pi->EIT){
        case EIT_Cursor: {
#define AddColor() { \
  u16 col=*pbuf; \
  col=((col&0x7BDE)>>1)+RGB15(14,14,16)|BIT15; \
  *pbuf=col; \
}
#define SetColor() { \
  u16 col=*pbuf; \
  col=((col&0x739C)>>2)+RGB15(22,22,24)|BIT15; \
  *pbuf=col; \
}
          const u32 x=pi->fx/0x100,y=pi->fy/0x100;
          if(50<LT){
            u16 *pbuf=&pcanbuf[((y-1)*ScreenWidth)+x];
            AddColor(); pbuf+=ScreenWidth-1;
            AddColor(); pbuf++;
            SetColor(); pbuf++;
            AddColor(); pbuf+=ScreenWidth-1;
            AddColor();
            }else{
            if(40<LT){
              u16 *pbuf=&pcanbuf[((y-1)*ScreenWidth)+x];
              AddColor(); pbuf+=ScreenWidth-1;
              AddColor(); pbuf++;
              AddColor(); pbuf++;
              AddColor(); pbuf+=ScreenWidth-1;
              AddColor();
              }else{
              if(30<LT){
                u16 *pbuf=&pcanbuf[((y-1)*ScreenWidth)+x];
                AddColor(); pbuf+=ScreenWidth-1;
                AddColor(); pbuf++;
                AddColor(); pbuf++;
                AddColor(); pbuf+=ScreenWidth-1;
                }else{
                if(20<LT){
                  u16 *pbuf=&pcanbuf[(y*ScreenWidth)+(x-1)];
                  AddColor(); pbuf++;
                  AddColor(); pbuf++;
                  AddColor(); pbuf+=ScreenWidth-1;
                  }else{
                  if(10<LT){
                    u16 *pbuf=&pcanbuf[(y*ScreenWidth)+(x-1)];
                    AddColor(); pbuf++;
                    AddColor(); pbuf++;
                    }else{
                    u16 *pbuf=&pcanbuf[(y*ScreenWidth)+x];
                    AddColor(); pbuf++;
                  }
                }
              }
            }
          }
#undef AddColor
#undef SetColor
        } break;
        case EIT_Fire: {
#define AddColor() { \
  u16 col=*pbuf; \
  col=((col&0x7BDE)>>1)+RGB15(4,4,16)|BIT15; \
  *pbuf=col; \
}
#define SetColor() { \
  u16 col=*pbuf; \
  col=((col&0x739C)>>2)+RGB15(6,6,24)|BIT15; \
  *pbuf=col; \
}
          const u32 x=pi->fx/0x100,y=pi->fy/0x100;
          if(30<LT){
            u16 *pbuf=&pcanbuf[((y-1)*ScreenWidth)+x];
            AddColor(); pbuf+=ScreenWidth-1;
            AddColor(); pbuf++;
            SetColor(); pbuf++;
            }else{
            u16 *pbuf=&pcanbuf[((y-1)*ScreenWidth)+x];
            AddColor(); pbuf+=ScreenWidth-1;
            AddColor(); pbuf++;
            AddColor(); pbuf++;
          }
#undef AddColor
#undef SetColor
        } break;
        case EIT_Bomb: {
#define AddColor() { \
  u16 col=*pbuf; \
  col=((col&0x7BDE)>>1)+RGB15(16,0,0)|BIT15; \
  *pbuf=col; \
}
#define SetColor() { \
  u16 col=*pbuf; \
  col=((col&0x739C)>>2)+RGB15(24,0,0)|BIT15; \
  *pbuf=col; \
}
          const u32 x=pi->fx/0x100,y=pi->fy/0x100;
          if(30<LT){
            u16 *pbuf=&pcanbuf[((y-1)*ScreenWidth)+x];
            AddColor(); pbuf+=ScreenWidth-1;
            AddColor(); pbuf++;
            SetColor(); pbuf++;
            }else{
            u16 *pbuf=&pcanbuf[((y-1)*ScreenWidth)+x];
            AddColor(); pbuf+=ScreenWidth-1;
            AddColor(); pbuf++;
            AddColor(); pbuf++;
          }
#undef AddColor
#undef SetColor
        } break;
      }
    }
  }
}

static int rand(int max)
{
  int ret=rand()/((double)RAND_MAX+1)*max;
  return(ret);
}

static u32 GetFreeIndex(void)
{
  u32 FreeIndex=(u32)-1;
  
  for(u32 idx=0;idx<ItemsCount;idx++){
    TItem *pi=&Items[idx];
    if(pi->LifeTime==0){
      FreeIndex=idx;
      break;
    }
  }
  
  if(FreeIndex==(u32)-1){
    u32 min=0xffffffff;
    for(u32 idx=0;idx<ItemsCount;idx++){
      TItem *pi=&Items[idx];
      if(pi->LifeTime<min) min=pi->LifeTime;
    }
    for(u32 idx=0;idx<ItemsCount;idx++){
      TItem *pi=&Items[idx];
      if(pi->LifeTime==min){
        FreeIndex=idx;
        break;
      }
    }
  }
  
  return(FreeIndex);
}

void Particle_SetCursor(s32 x,s32 y,s32 vx,s32 vy)
{
// Already enabled.
//  if(Enabled==false) return;
  
  if(ProcState.Main.UseScreenEffect==false) return;
  
  if((x<0)||(ScreenWidth<x)) return;
  if((y<0)||(ScreenHeight<y)) return;
  
  u32 FreeIndex=GetFreeIndex();
  if(FreeIndex==(u32)-1) return;
  
  TItem *pi=&Items[FreeIndex];
  pi->LifeTime=60+rand(60);
  pi->EIT=EIT_Cursor;
  pi->fx=x*0x100+rand(4*0x100)-(2*0x100);
  pi->fy=y*0x100+rand(4*0x100)-(2*0x100);
  pi->fvx=(-vx*0x8)+rand(1*0x100)-(0.5*0x100);
  pi->fvy=(-vy*0x8)+rand(0.5*0x100)-(0.5*0x100);
}

void Particle_SetFire(s32 x,s32 y)
{
  if(Enabled==false) return;
  
  if((x<0)||(ScreenWidth<x)) return;
  if((y<0)||(ScreenHeight<y)) return;
  
  for(u32 idx=0;idx<8;idx++){
    u32 FreeIndex=GetFreeIndex();
    if(FreeIndex==(u32)-1) return;
    
    TItem *pi=&Items[FreeIndex];
    pi->LifeTime=30+rand(30);
    pi->EIT=EIT_Fire;
    pi->fx=x*0x100+rand(16*0x100)-(8*0x100);
    pi->fy=y*0x100+rand(16*0x100)-(8*0x100);
    pi->fvx=rand(1*0x100)-(0.5*0x100);
    pi->fvy=rand(0.5*0x100)-(0.5*0x100);
  }
}

void Particle_SetFireSnow(s32 x,s32 y)
{
  if(Enabled==false) return;
  
  if((x<0)||(ScreenWidth<x)) return;
  if((y<0)||(ScreenHeight<y)) return;
  
  for(u32 idx=0;idx<1;idx++){
    u32 FreeIndex=GetFreeIndex();
    if(FreeIndex==(u32)-1) return;
    
    TItem *pi=&Items[FreeIndex];
    pi->LifeTime=30+rand(30);
    pi->EIT=EIT_Fire;
    pi->fx=x*0x100+rand(16*0x100)-(8*0x100);
    pi->fy=y*0x100+rand(16*0x100)-(8*0x100);
    pi->fvx=rand(1*0x100)-(0.5*0x100);
    pi->fvy=rand(0.5*0x100)-(0.5*0x100);
  }
}

void Particle_SetFireBomb(s32 x,s32 y)
{
  if(Enabled==false) return;
  
  if((x<0)||(ScreenWidth<x)) return;
  if((y<0)||(ScreenHeight<y)) return;
  
  for(u32 idx=0;idx<256;idx++){
    u32 FreeIndex=GetFreeIndex();
    if(FreeIndex==(u32)-1) return;
    
    TItem *pi=&Items[FreeIndex];
    pi->LifeTime=120+rand(120);
    pi->EIT=EIT_Bomb;
    pi->fx=x*0x100+rand(64*0x100)-(32*0x100);
    pi->fy=y*0x100+rand(64*0x100)-(32*0x100);
    pi->fvx=rand(8*0x100)-(4*0x100);
    pi->fvy=rand(6*0x100)-(4*0x100);
  }
}

extern void Particle_SetFireAutoFlag(s32 x,s32 y)
{
  if(Enabled==false) return;
  
  if((x<0)||(ScreenWidth<x)) return;
  if((y<0)||(ScreenHeight<y)) return;
  
  for(u32 idx=0;idx<8;idx++){
    u32 FreeIndex=GetFreeIndex();
    if(FreeIndex==(u32)-1) return;
    
    TItem *pi=&Items[FreeIndex];
    pi->LifeTime=30+rand(30);
    pi->EIT=EIT_Fire;
    pi->fx=x*0x100+rand(16*0x100)-(8*0x100);
    pi->fy=y*0x100+rand(16*0x100)-(8*0x100);
    pi->fvx=rand(1*0x100)-(0.5*0x100);
    pi->fvy=rand(0.5*0x100)-(0.5*0x100);
  }
}

void Particle_Move(s32 x,s32 y)
{
  if(Enabled==false) return;
  
  s32 fx=x*0x100,fy=y*0x100;
  
  for(u32 idx=0;idx<ItemsCount;idx++){
    TItem *pi=&Items[idx];
    if(pi->EIT!=EIT_Cursor){
      if(pi->LifeTime!=0){
        pi->fx-=fx;
        pi->fy-=fy;
      }
    }
  }
}

