
#include <nds.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "_console.h"
#include "_consolewritelog.h"
#include "maindef.h"
#include "memtool.h"
#include "_const.h"
#include "lang.h"
#include "datetime.h"
#include "inifile.h"

#include "glib.h"
#include "rect.h"
#include "skin.h"
#include "skinpack.h"
#include "strtool.h"

#include "fat2.h"
#include "shell.h"
#include "sndeff.h"
#include "procstate.h"
#include "strpcm.h"
#include "cfont.h"
#include "lngmsg.h"

#include "libms.h"
#include "Particle.h"

static int rand(int max)
{
  int ret=rand()/((double)RAND_MAX+1)*max;
  return(ret);
}

static u32 EndKeyState;

static u32 CurLevel;

static CglTGF *pBackBtn;

// ------------------------------------------------------

static void StartBGM(void)
{
  EWAVFN WAVFNs[16];
  u32 WAVFNsCount=0;
  if(ProcState.Main.BGMSelectHouse==true){
    WAVFNs[WAVFNsCount++]=WAVFN_BGMHouse1;
    WAVFNs[WAVFNsCount++]=WAVFN_BGMHouse2;
    WAVFNs[WAVFNsCount++]=WAVFN_BGMHouse3;
  }
  if(ProcState.Main.BGMSelectJazz==true){
    WAVFNs[WAVFNsCount++]=WAVFN_BGMJazz1;
  }
  if(ProcState.Main.BGMSelectMar==true){
    WAVFNs[WAVFNsCount++]=WAVFN_BGMMar1;
  }
  if(ProcState.Main.BGMSelectRB==true){
    WAVFNs[WAVFNsCount++]=WAVFN_BGMRB1;
    WAVFNs[WAVFNsCount++]=WAVFN_BGMRB2;
  }
  SoundBGM_Start(WAVFNs[rand(WAVFNsCount)],true);
}

static void EndBGM(void)
{
  SoundBGM_Stop();
}

// ------------------------------------------------------

static CglTGF *pHelpScreen_BG,*pHelpScreen_Btn;
static u32 HelpScreenTimeout;
static u32 HelpScreenShowBtn;

static void HelpScreen_Init(void)
{
  pHelpScreen_BG=SkinPack_LoadTGF(ESF_GameMainHelpBG_tgf);
  pHelpScreen_Btn=SkinPack_LoadTGF(ESF_GameMainHelpBtn_tgf);
  
  HelpScreenTimeout=0;
  HelpScreenShowBtn=false;
}

static void HelpScreen_ShowBtn(void)
{
  HelpScreenShowBtn=true;
}

static void HelpScreen_HideBtn(void)
{
  HelpScreenShowBtn=false;
}

static void HelpScreen_Free(void)
{
  if(pHelpScreen_BG!=NULL){
    delete pHelpScreen_BG; pHelpScreen_BG=NULL;
  }
  if(pHelpScreen_Btn!=NULL){
    delete pHelpScreen_Btn; pHelpScreen_Btn=NULL;
  }
}

static bool HelpScreen_isInside(u32 x,u32 y)
{
  if(HelpScreenShowBtn==false) return(false);
  
  const u32 w=pHelpScreen_Btn->GetWidth(),h=pHelpScreen_Btn->GetHeight();
  if(((ScreenWidth-w)<=x)&&((ScreenHeight-h)<=y)) return(true);
  return(false);
}

static void HelpScreen_Show(void)
{
  HelpScreenShowBtn=false;
  
  if(HelpScreenTimeout!=0){
    HelpScreenTimeout=0;
    return;
  }
  
  StartBGM();
  HelpScreenTimeout=5*60;
}

static void HelpScreen_Draw(CglCanvas *pcan)
{
  if(HelpScreenShowBtn==true){
    const u32 w=pHelpScreen_Btn->GetWidth(),h=pHelpScreen_Btn->GetHeight();
    pHelpScreen_Btn->BitBlt(pcan,ScreenWidth-w,ScreenHeight-h);
  }
  
  if(HelpScreenTimeout!=0){
    HelpScreenTimeout--;
    pHelpScreen_BG->BitBlt(pcan,0,0);
  }
}


// ------------------------------------------------------

static char SysMsg_str[256];
static u32 SysMsg_Timeout;

static void SysMsg_Init(void)
{
  SysMsg_str[0]=0;
  SysMsg_Timeout=0;
}

static void SysMsg_SetMsgLang(const char *pstr)
{
  strcpy(SysMsg_str,pstr);
  SysMsg_Timeout=2*60;
  SetProcFadeEffect(EPFE_CrossFade);
}

static void SysMsg_UpdateVsync(const u32 VsyncCount)
{
  if(SysMsg_Timeout==0) return;
  
  if(MS_GetGameMode()!=EGM_Play) return;
  
  if(SysMsg_Timeout<VsyncCount){
    SysMsg_Timeout=0;
    }else{
    SysMsg_Timeout-=VsyncCount;
  }
  
  if(SysMsg_Timeout==0) SetProcFadeEffect(EPFE_FastCrossFade);
}

static void SysMsg_Draw(CglCanvas *pcan)
{
  if(SysMsg_Timeout==0) return;
  
  const u32 tw=pcan->GetTextWidthLang(SysMsg_str);
  const u32 tx=(ScreenWidth-tw)/2;
  const u32 th=glCanvasTextHeight+8;
  const u32 ty=((ScreenHeight-th)/2);
  
  {
    u16 *pbuf=pcan->GetVRAMBuf();
    pbuf+=(ty-1)*ScreenWidth;
    for(u32 idx=0;idx<1*ScreenWidth;idx++){
      u16 col=*pbuf;
      col=((col&0x7BDE)>>1)+RGB15(0,0,0)|BIT15;
      *pbuf++=col;
    }
    for(u32 idx=0;idx<th*ScreenWidth;idx++){
      u16 col=*pbuf;
      col=((col&0x739C)>>2)+RGB15(22,20,20)|BIT15;
      *pbuf++=col;
    }
    for(u32 idx=0;idx<1*ScreenWidth;idx++){
      u16 col=*pbuf;
      col=((col&0x7BDE)>>1)+RGB15(0,0,0)|BIT15;
      *pbuf++=col;
    }
  }
  
  pcan->SetFontTextColor(RGB15(0,0,0)|BIT15);
  pcan->TextOutLang(tx,ty+4,SysMsg_str);
}

// ------------------------------------------------------

typedef struct {
  float x,y;
} TFPoint;
static TFPoint CurPos,TagPos;

static void CB_KeyPress(u32 VsyncCount,u32 Keys,bool FirstFlag)
{
  if((Keys&KEY_L)!=0){
    if((Keys&(KEY_X|KEY_Y))!=0){
      if((Keys&KEY_X)!=0) ChangeNextBacklightLevel();
      if((Keys&KEY_Y)!=0) ChangePrevBacklightLevel();
      return;
    }
    return;
  }
  
  if((Keys&KEY_SELECT)!=0){
    HelpScreen_Show();
  }
  
  if((Keys&KEY_START)!=0){
    MS_Interrupt();
  }
  
  if((Keys&(KEY_X|KEY_Y))!=0){
    SoundSE_Start(WAVFN_Zoom);
    if((Keys&KEY_X)!=0) MS_BtmScrZoom();
    if((Keys&KEY_Y)!=0) MS_BtmScrReduce();
    TPoint pos=MS_GetScroll();
    TagPos=(TFPoint){pos.x,pos.y};
    CurPos=TagPos;
  }
  
}

static bool MouseDownFlag;
static s32 mx,my;
static s32 msx,msy;
static s32 MousePressButton;
static u32 MouseDownVsyncCount;
static TPoint MouseFirstTouchCell;
static bool MouseIgnoreOpenCell;

static void CB_MouseDown(s32 x,s32 y)
{
  MouseDownFlag=false;
  
  if(MS_GetGameMode()!=EGM_Play){
    if(((ScreenWidth-pBackBtn->GetWidth())<=x)&&(ScreenHeight-pBackBtn->GetHeight())<=y){
      return;
    }
  }
  
  if(HelpScreen_isInside(x,y)==true){
    HelpScreen_Show();
    return;
  }
  
  MouseDownFlag=true;
  MouseDownVsyncCount=0;
  
  msx=x; msy=y;
  
  MousePressButton=false;
  if(MS_GetGameMode()==EGM_Play){
    u32 keys=(~REG_KEYINPUT)&0x3ff;
    if(keys&(KEY_L|KEY_R|KEY_UP|KEY_DOWN|KEY_LEFT|KEY_RIGHT|KEY_A|KEY_B)) MousePressButton=true;
  }
  
  if(MousePressButton==true){
    MouseFirstTouchCell=MS_GetCellIndexFromScreenView(x,y);
    MS_SetPreview(MouseFirstTouchCell.x,MouseFirstTouchCell.y);
    MouseIgnoreOpenCell=false;
  }
  
  mx=x;
  my=y;
}

static void CB_Mouse_PressButtonCancel(void)
{
  MS_UnsetPreview();
  MouseIgnoreOpenCell=true;
}

static void CB_MouseMove(s32 x,s32 y)
{
  if(MouseDownFlag==false) return;
  
  if(MousePressButton==false){
    const s32 vx=mx-x,vy=my-y;
    
    const s32 cnt=sqrt((float)((vx*vx)+(vy*vy)));
    for(s32 idx=0;idx<cnt;idx++){
      u32 px=mx-(vx*idx/cnt),py=my-(vy*idx/cnt);
      Particle_SetCursor(px,py,vx,vy);
    }
    
    TagPos.x+=vx;
    TagPos.y+=vy;
    }else{
    TPoint CellNum=MS_GetCellIndexFromScreenView(x,y);
    if((MouseFirstTouchCell.x!=CellNum.x)||(MouseFirstTouchCell.y!=CellNum.y)) CB_Mouse_PressButtonCancel();
  }
  
  mx=x;
  my=y;
}

static void CB_MouseUp(s32 x,s32 y)
{
  if(MS_GetGameMode()!=EGM_Play){
    if(((ScreenWidth-pBackBtn->GetWidth())<=x)&&(ScreenHeight-pBackBtn->GetHeight())<=y){
      SetNextProc(ENP_Title,EPFE_None);
      return;
    }
  }
  
  if(MouseDownFlag==false) return;
  MouseDownFlag=false;
  
  if(MousePressButton==false){
    TPoint pos={TagPos.x,TagPos.y};
    pos=MS_InsideScreen(pos);
    TagPos=(TFPoint){pos.x,pos.y};
    if((MouseDownVsyncCount<8)&&((abs(msx-x)+abs(msy-y))<8)){
      TPoint CellNum=MS_GetCellIndexFromScreenView(mx,my);
      MS_ChangeCheck(CellNum.x,CellNum.y);
    }
    }else{
    bool ig=MouseIgnoreOpenCell;
    MS_UnsetPreview();
    if(ig==false) MS_OpenCell(MouseFirstTouchCell.x,MouseFirstTouchCell.y);
  }
}

static void CB_Start(void)
{
  EndKeyState=0;
  
  MouseDownFlag=false;
  CurPos=(TFPoint){0,0};
  TagPos=(TFPoint){0,0};
  
  StartBGM();
  
  pBackBtn=SkinPack_LoadTGF(ESF_GameMainBackBtn_tgf);
  
  SysMsg_Init();
  
  Particle_Init(true);
  
  CurLevel=ProcState.Main.PlayLevel;
  
  HelpScreen_Init();
  
  if(CurLevel==1){
    HelpScreen_ShowBtn();
    SysMsg_SetMsgLang(LNGMSG_GameMain_SelectButtonShowHelp);
  }
  
  u32 cx,cy,bomb;
  switch(CurLevel){
    case 0: cx=ProcState.Main.CustomWidth; cy=ProcState.Main.CustomHeight; bomb=ProcState.Main.CustomBombs; break;
    case 1: cx=10; cy=7; bomb=5; break;
    case 2: cx=17; cy=12; bomb=30; break;
    case 3: cx=24; cy=17; bomb=70; break;
    case 4: cx=30; cy=22; bomb=120; break;
    case 5: cx=30; cy=22; bomb=200; break;
    default: {
      StopFatalError(0,"Undefined level %d.",CurLevel);
    } break;
  }
//  if(CurLevel!=1){ cx-=4;cy-=4;bomb/=4; }
  MS_Init(CurLevel,cx+2,cy+2,bomb,true,ProcState.Main.CellSizeBtm);
  MS_SetEnableSE(true);
  MS_Refresh();
}

static void CB_VsyncUpdate(u32 VsyncCount)
{
  if(IPC6->PanelOpened==false) return;
  
  if(MouseDownFlag==true){
    MouseDownVsyncCount+=VsyncCount;
    if(16<=MouseDownVsyncCount) CB_Mouse_PressButtonCancel();
  }
  MS_VsyncInterrupt(VsyncCount);
  
  if(4<VsyncCount) VsyncCount=4;
  SysMsg_UpdateVsync(VsyncCount);
  
  if(VsyncCount<4){
    static u32 ProcState_SaveInterval=0;
    if(ProcState_SaveInterval<=2){
      ProcState_SaveInterval++;
      }else{
      ProcState_SaveInterval=0;
      ProcState_Save();
    }
  }
  
  for(u32 idx=0;idx<VsyncCount;idx++){
    float vx=(TagPos.x-CurPos.x)/4;
    float vy=(TagPos.y-CurPos.y)/4;
    CurPos.x+=vx;
    CurPos.y+=vy;
    const TPoint pos={CurPos.x,CurPos.y};
    MS_SetScroll(pos);
    Particle_Move(vx,vy);
  }
  
  MS_ProcessAutoOpen();
  
  if(MS_GetGameMode()==EGM_Play){
    s32 flags=MS_GetRemainFlagsCount();
    static s32 lastflags=0;
    if((flags==0)&&(0<lastflags)){
      SoundSE_Start(WAVFN_Setting);
      SysMsg_SetMsgLang(LNGMSG_GameMain_FlagError);
    }
    lastflags=flags;
  }
  
  if(MS_GetGameMode()==EGM_Clear){
    u32 x=rand(ScreenWidth),y=rand(ScreenHeight);
    for(u32 idx=0;idx<8;idx++){
      Particle_SetCursor(x,y,0,0);
    }
  }
  
  for(u32 idx=0;idx<VsyncCount;idx++){
    Particle_Process();
  }
  
  MS_Refresh();
  
  {
    bool draw=false;
    static u32 a=0;
    if(MouseDownFlag==false){
      if(a!=0){
        a--;
        }else{
        a=3;
        draw=true;
      }
      }else{
      if(a!=0){
        a--;
        }else{
        a=6;
        draw=true;
      }
    }
    if(draw==true) MS_Draw_Top(pScreenSub->pCanvas,true);
  }
  
  CglCanvas *pcan=pScreenMain->pBackCanvas;
  pcan->SetCglFont(pCglFontDefault);
  
  MS_Draw_Btm(pcan);
  Particle_Draw(pcan);
  SysMsg_Draw(pcan);
  
  HelpScreen_Draw(pcan);
  
  if(MS_GetGameMode()!=EGM_Play){
    pBackBtn->BitBlt(pcan,ScreenWidth-pBackBtn->GetWidth(),ScreenHeight-pBackBtn->GetHeight());
  }
  
  ScreenMain_Flip_ProcFadeEffect();
  
  if(MS_GetGameMode()!=EGM_Play){
    HelpScreen_HideBtn();
    u32 keys=(~REG_KEYINPUT)&0x3ff;
    switch(EndKeyState){
      case 0: {
        EndKeyState=1;
        switch(MS_GetGameMode()){
          case EGM_Play: {
          } break;
          case EGM_Clear: {
            EndBGM();
            if(rand(2)==0){
              SoundSE_Start(WAVFN_Clear1);
              }else{
              SoundSE_Start(WAVFN_Clear2);
            }
            if(CurLevel==0){
              SysMsg_SetMsgLang(LNGMSG_GameMain_Clear_Custom);
              }else{
              if(CurLevel==5){
                SysMsg_SetMsgLang(LNGMSG_GameMain_Clear_LevelX);
                }else{
                SysMsg_SetMsgLang(LNGMSG_GameMain_Clear_Normal);
              }
            }
          } break;
          case EGM_GameOver: {
            EndBGM();
            SoundSE_Start(WAVFN_Bomb);
            char msg[128];
            const u32 RestBomb=MS_GetRestBomb();
            if(CurLevel==5){
              snprintf(msg,128,LNGMSG_GameMain_GameOver_LevelX,RestBomb,RestBomb);
              }else{
              snprintf(msg,128,LNGMSG_GameMain_GameOver_Normal,RestBomb,RestBomb);
            }
            SysMsg_SetMsgLang(msg);
          } break;
          case EGM_Interrupt: {
            SoundSE_Start(WAVFN_GiveUp);
            if(MS_GetBombsSetupped()==false){
              SetNextProc(ENP_Title,EPFE_CrossFade);
              }else{
              MS_ShowAllBombs();
              SysMsg_SetMsgLang(LNGMSG_GameMain_Interrupt);
            }
          } break;
        }
      } break;
      case 1: {
        if(keys==0) EndKeyState=2;
      } break;
      case 2: {
        if(keys!=0) EndKeyState=3;
      } break;
      case 3: {
        if(keys==0) SetNextProc(ENP_Title,EPFE_None);
      } break;
    }
  }
}

static void CB_End(void)
{
  const u32 Level=ProcState.Main.PlayLevel;
  if(Level<ProcState_Score_Count){
    switch(MS_GetGameMode()){
      case EGM_Play: {
      } break;
      case EGM_Clear: {
        ProcState.Score.PlayCount[Level]++;
        ProcState.Score.ClearCount[Level]++;
        u32 time=MS_GetElapsedTimeTime();
        u32 fasttime=ProcState.Score.FastTime[Level];
        if((fasttime==0)||(time<fasttime)){
          fasttime=time;
          ProcState.Score.FastTime[Level]=fasttime;
          _consolePrintf("Level.%d registed highscore. %fsecs.\n",Level,(float)fasttime/60);
        }
      } break;
      case EGM_GameOver: {
        ProcState.Score.PlayCount[Level]++;
        u32 RestBomb=MS_GetRestBomb();
        u32 back=ProcState.Score.RestBomb[Level];
        _consolePrintf("%d,%d\n",RestBomb,back);
        if((back==0)||(RestBomb<back)){
          back=RestBomb;
          ProcState.Score.RestBomb[Level]=back;
          _consolePrintf("Level.%d registed highscore. %dbombs.\n",Level,back);
        }
      } break;
      case EGM_Interrupt: {
      } break;
    }
  }
  
  if((Level==1)&&(ProcState.Score.PlayCount[1]!=0)&&(ProcState.Score.ClearCount[1]==1)) TitleShowHighScoreMsgBox=true;
  
  ProcState.Main.CellSizeBtm=MS_GetBottomCellSize();
  
  ProcState_RequestSave=true;
  ProcState_Save();
  
  SoundSE_Stop();
  
  EndBGM();
  
  Particle_Free();
  
  HelpScreen_Free();
  
  MS_Free();
  
  if(pBackBtn!=NULL){
    delete pBackBtn; pBackBtn=NULL;
  }
}

static void CB_ReqSoftReset(void)
{
  pScreenSub->pCanvas->FillFull(0);
  
  pScreenMain->pBackCanvas->FillFull(RGB15(0,0,0)|BIT15);
  SetProcFadeEffect(EPFE_UpToDown);
  ScreenMain_Flip_ProcFadeEffect();
  pScreenMain->pBackCanvas->FillFull(RGB15(0,0,0)|BIT15);
  
  extern bool ReqSoftResetFlag;
  ReqSoftResetFlag=true;
}

static void CB_PanelOpen(void)
{
//  SoundSE_Start(WAVFN_PanelOpen);
}

static void CB_PanelClose(void)
{
  SoundSE_Start(WAVFN_PanelClose);
}

void ProcMain_SetCallBack(TCallBack *pCallBack)
{
  pCallBack->Start=CB_Start;
  pCallBack->VsyncUpdate=CB_VsyncUpdate;
  pCallBack->End=CB_End;
  pCallBack->KeyPress=CB_KeyPress;
  pCallBack->MouseDown=CB_MouseDown;
  pCallBack->MouseMove=CB_MouseMove;
  pCallBack->MouseUp=CB_MouseUp;
  pCallBack->ReqSoftReset=CB_ReqSoftReset;
  pCallBack->PanelOpen=CB_PanelOpen;
  pCallBack->PanelClose=CB_PanelClose;
}

