
#include <nds.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "_console.h"
#include "_consolewritelog.h"
#include "maindef.h"
#include "memtool.h"
#include "_const.h"
#include "lang.h"
#include "datetime.h"
#include "inifile.h"

#include "glib.h"
#include "skin.h"
#include "skinpack.h"
#include "strtool.h"

#include "fat2.h"
#include "shell.h"
#include "sndeff.h"
#include "procstate.h"
#include "strpcm.h"
#include "cfont.h"
#include "rect.h"

#include "libnumbox.h"

// ----------------------------------------------------------------

static CglTGF *pNumBox_D2OpenBG,*pNumBox_D2ClosedBG;
static CglTGF *pNumBox_D3OpenBG,*pNumBox_D3ClosedBG;
static CglTGF *pNumBox_Font;
static CglTGF *pNumBox_OKBtn;

void NumBox_Init(void)
{
  pNumBox_D2OpenBG=SkinPack_LoadTGF(ESF_NumBox_D2OpenBG_tgf);
  pNumBox_D2ClosedBG=SkinPack_LoadTGF(ESF_NumBox_D2ClosedBG_tgf);
  pNumBox_D3OpenBG=SkinPack_LoadTGF(ESF_NumBox_D3OpenBG_tgf);
  pNumBox_D3ClosedBG=SkinPack_LoadTGF(ESF_NumBox_D3ClosedBG_tgf);
  pNumBox_Font=SkinPack_LoadTGF(ESF_NumBox_Font_tgf);
  pNumBox_OKBtn=SkinPack_LoadTGF(ESF_NumBox_OKBtn_tgf);
}

void NumBox_Free(void)
{
  if(pNumBox_D2OpenBG!=NULL){
    delete pNumBox_D2OpenBG; pNumBox_D2OpenBG=NULL;
  }
  if(pNumBox_D2ClosedBG!=NULL){
    delete pNumBox_D2ClosedBG; pNumBox_D2ClosedBG=NULL;
  }
  if(pNumBox_D3OpenBG!=NULL){
    delete pNumBox_D3OpenBG; pNumBox_D3OpenBG=NULL;
  }
  if(pNumBox_D3ClosedBG!=NULL){
    delete pNumBox_D3ClosedBG; pNumBox_D3ClosedBG=NULL;
  }
  if(pNumBox_Font!=NULL){
    delete pNumBox_Font; pNumBox_Font=NULL;
  }
  if(pNumBox_OKBtn!=NULL){
    delete pNumBox_OKBtn; pNumBox_OKBtn=NULL;
  }
}

CglTGF* NumBox_GetFont(void)
{
  return(pNumBox_Font);
}

// ----------------------------------------------------------------

static const CglTGF* GetOpenBG(u32 Digits)
{
  switch(Digits){
    case 2: return(pNumBox_D2OpenBG); break;
    case 3: return(pNumBox_D3OpenBG); break;
  }
  
  StopFatalError(0,"Illigal digits. (%d)",Digits);
  return(NULL);
}

static const CglTGF* GetClosedBG(u32 Digits)
{
  switch(Digits){
    case 2: return(pNumBox_D2ClosedBG); break;
    case 3: return(pNumBox_D3ClosedBG); break;
  }
  
  StopFatalError(0,"Illigal digits. (%d)",Digits);
  return(NULL);
}

static TRect GetClosedBGRect(const CglTGF *ptgf,s32 x,s32 y)
{
  const s32 h=pNumBox_Font->GetHeight()/10;
  
  TRect r;
  r.w=ptgf->GetWidth();
  r.h=ptgf->GetHeight();
  r.x=x-(r.w/2);
  r.y=y-(r.h/2)+(h/2);
  return(r);
}

static TRect GetOpenBGRect(const CglTGF *ptgf,s32 x,s32 y)
{
  const s32 h=pNumBox_Font->GetHeight()/10;
  
  TRect r;
  r.w=ptgf->GetWidth();
  r.h=ptgf->GetHeight();
  r.x=x-(r.w/2);
  r.y=y-(r.h/2)+(h/2);
  return(r);
}

CNumBox::CNumBox(const u32 _Digits,const s32 _PosLeft,const s32 _PosTop,const u32 DefaultNum)
                : Digits(_Digits),
                  pOpenBG(GetOpenBG(Digits)),OpenBGRect(GetOpenBGRect(pOpenBG,_PosLeft,_PosTop)),
                  pClosedBG(GetClosedBG(Digits)),ClosedBGRect(GetClosedBGRect(pClosedBG,_PosLeft,_PosTop)),
                  pFont(pNumBox_Font),FontWidth(pNumBox_Font->GetWidth()),FontHeight(pNumBox_Font->GetHeight()/10),
                  pOKBtn(pNumBox_OKBtn),
                  PosLeft(_PosLeft),PosTop(_PosTop)
{
  Enabled=false;
  
  {
    const TRect r=OpenBGRect;
    pWindowBG=new CglCanvas(&MM_Process,NULL,r.w,r.h,pf15bit);
  }

  Min=0;
  Max=0;
  
  CurY=DefaultNum*FontHeight;
  TagY=DefaultNum*FontHeight;
  
  mf=false;
  mx=0;
  my=0;
  mv=0;
  
  for(u32 idx=0;idx<3;idx++){
    MouseLogs[idx].mx=0;
    MouseLogs[idx].my=0;
    MouseLogs[idx].mv=60;
  }
}

CNumBox::~CNumBox(void)
{
  Enabled=false;
  
  if(pWindowBG!=NULL){
    delete pWindowBG; pWindowBG=NULL;
  }
}

void CNumBox::SetLimit(s32 _Min,s32 _Max)
{
  Min=_Min;
  Max=_Max;
}

void CNumBox::SetCurrentNumber(s32 num)
{
  CurY=num*FontHeight;
  TagY=CurY;
}

s32 CNumBox::GetCurrentNumber(void)
{
  return(CurY/FontHeight);
}

bool CNumBox::NumClosedBG_isInside(s32 x,s32 y)
{
  TRect r=ClosedBGRect;
  if((r.x<=x)&&(x<(r.x+r.w))&&(r.y<=y)&&(y<(r.y+r.h))) return(true);
  return(false);
}

void CNumBox::Disabled(void)
{
  if(TagY<(Min*FontHeight)) TagY=Min*FontHeight;
  if((Max*FontHeight)<TagY) TagY=Max*FontHeight;
  
  CurY=TagY;
  Enabled=false;
  
  SoundSE_Start(WAVFN_Open);
  SetProcFadeEffect(EPFE_FastCrossFade);
}

void CNumBox::StoreMouseLog(u32 mx,u32 my,u32 mv)
{
  MouseLogs[0]=MouseLogs[1];
  MouseLogs[1]=MouseLogs[2];
  MouseLogs[2].mx=mx;
  MouseLogs[2].my=my;
  MouseLogs[2].mv=mv;
}

static void gausscolor50bk(CglCanvas *pcan)
{
  CglCanvas *psrccan=new CglCanvas(&MM_Temp,NULL,ScreenWidth,ScreenHeight,pf15bit);
  pcan->BitBltFullBeta(psrccan);
  
  CglCanvas *pdstcan=new CglCanvas(&MM_Temp,NULL,ScreenWidth,ScreenHeight,pf15bit);
  pdstcan->FillFull(RGB15(0,0,0)|BIT15);
  
    u32 gr=0,gg=0,gb=0;
  for(u32 y=1;y<(ScreenHeight-1);y++){
    u16 *psrcbufm1=psrccan->GetScanLine(y-1);
    u16 *psrcbuf=psrccan->GetScanLine(y);
    u16 *psrcbufp1=psrccan->GetScanLine(y+1);
    u16 *ptagbuf=pdstcan->GetScanLine(y);
    for(u32 x=1;x<(ScreenWidth-1);x++){
      u32 c0=psrcbuf[x];
      u32 c1=psrcbufm1[x];
      u32 c2=psrcbuf[x-1];
      u32 c3=psrcbuf[x+1];
      u32 c4=psrcbufp1[x];
      u32 c5=psrcbufm1[x-1];
      u32 c6=psrcbufm1[x+1];
      u32 c7=psrcbufp1[x-1];
      u32 c8=psrcbufp1[x+1];
      u32 r0=(c0&(0x1f<<0))>>0;
      u32 r1=(c1&(0x1f<<0))>>0;
      u32 r2=(c2&(0x1f<<0))>>0;
      u32 r3=(c3&(0x1f<<0))>>0;
      u32 r4=(c4&(0x1f<<0))>>0;
      u32 r5=(c5&(0x1f<<0))>>0;
      u32 r6=(c6&(0x1f<<0))>>0;
      u32 r7=(c7&(0x1f<<0))>>0;
      u32 r8=(c8&(0x1f<<0))>>0;
      u32 g0=(c0&(0x1f<<5))>>5;
      u32 g1=(c1&(0x1f<<5))>>5;
      u32 g2=(c2&(0x1f<<5))>>5;
      u32 g3=(c3&(0x1f<<5))>>5;
      u32 g4=(c4&(0x1f<<5))>>5;
      u32 g5=(c5&(0x1f<<5))>>5;
      u32 g6=(c6&(0x1f<<5))>>5;
      u32 g7=(c7&(0x1f<<5))>>5;
      u32 g8=(c8&(0x1f<<5))>>5;
      u32 b0=(c0&(0x1f<<10))>>10;
      u32 b1=(c1&(0x1f<<10))>>10;
      u32 b2=(c2&(0x1f<<10))>>10;
      u32 b3=(c3&(0x1f<<10))>>10;
      u32 b4=(c4&(0x1f<<10))>>10;
      u32 b5=(c5&(0x1f<<10))>>10;
      u32 b6=(c6&(0x1f<<10))>>10;
      u32 b7=(c7&(0x1f<<10))>>10;
      u32 b8=(c8&(0x1f<<10))>>10;
      u32 r=((r0*4)+((r1+r2+r3+r4)*2)+((r5+r6+r7+r8)*1));
      u32 g=((g0*4)+((g1+g2+g3+g4)*2)+((g5+g6+g7+g8)*1));
      u32 b=((b0*4)+((b1+b2+b3+b4)*2)+((b5+b6+b7+b8)*1));
      const u32 lim=(32*32)-1;
//      r+=gr; gr=r&31; if(lim<r) r=lim;
//      g+=gg; gg=g&31; if(lim<g) g=lim;
//      b+=gb; gb=b&31; if(lim<b) b=lim;
      ptagbuf[x]=RGB15(r/32,g/32,b/32)|BIT15;
    }
  }
  
  pdstcan->BitBltFullBeta(pcan);
  
  delete psrccan; psrccan=NULL;
  delete pdstcan; pdstcan=NULL;
}

static bool IgnoreOKButton;

bool CNumBox::CB_MouseDown(s32 x,s32 y)
{
  StoreMouseLog(x,y,mv);
  
  mf=true;
  
  mx=x;
  my=y;
  mv=0;
  
  IgnoreOKButton=false;
  
  if(Enabled==false){
    if(NumClosedBG_isInside(x,y)==false){
      mf=false;
      return(false);
    }
    Enabled=true;
    
    CglCanvas *psrccan=pScreenMain->pViewCanvas;
    CglCanvas *pcan=pScreenMain->pBackCanvas;
    
    if(1){
      psrccan->BitBltFullBeta(pcan);
      gausscolor50bk(pcan);
      }else{
      u16 *psrcbuf=psrccan->GetVRAMBuf();
      u16 *pdstbuf=pcan->GetVRAMBuf();
      for(u32 idx=0;idx<ScreenWidth*ScreenHeight;idx++){
        u16 col=*psrcbuf++;
        col=(col&0x7BDE)>>1;
        col=(col&0x7BDE)>>1;
        col+=RGB15(0,0,0);
        *pdstbuf++=col|BIT15;
      }
    }
    
    const TRect r=OpenBGRect;
    pOpenBG->BitBlt(pcan,r.x,r.y);
    pcan->BitBlt(pWindowBG,0,0,r.w,r.h,r.x,r.y,false);
    
    pOKBtn->BitBlt(pcan,ScreenWidth-pOKBtn->GetWidth(),ScreenHeight-pOKBtn->GetHeight());
    
    Draw(pcan);
    
    SoundSE_Start(WAVFN_Open);
    SetProcFadeEffect(EPFE_FastCrossFade);
    ScreenMain_Flip_ProcFadeEffect();
    
    return(true);
  }
  
  if(((ScreenWidth-pOKBtn->GetWidth())<=x)&&((ScreenHeight-pOKBtn->GetHeight())<=y)){
    }else{
    IgnoreOKButton=true;
  }
  
  return(false);
}

void CNumBox::CB_MouseMove(s32 x,s32 y)
{
  if(mf==false) return;
  
  s32 Mul=0;
  switch(Digits){
    case 2: Mul=3; break;
    case 3: Mul=10; break;
  }
  
  TagY+=(my-y)*Mul;
  
  mx=x;
  my=y;
}

void CNumBox::CB_MouseUp(s32 x,s32 y)
{
  if(mf==false) return;
  mf=false;
  
  if(IgnoreOKButton==false){
    if(mv<16){
      if(((ScreenWidth-pOKBtn->GetWidth())<=x)&&((ScreenHeight-pOKBtn->GetHeight())<=y)){
        Disabled();
        return;
      }
    }
  }
  
  const s32 bgx=PosLeft,bgy=PosTop,bgh=FontHeight;
  s32 bgw=FontWidth;
  if(Digits==3) bgw*=1.5;
  bool IgnoreDoubleClick=false;
  
  if(mv<16){
    if(((bgx-bgw)<x)&&(x<=(bgx+bgw))){
      s32 num=(TagY+(bgh/2))/bgh;
      if(((bgy-bgh)<y)&&(y<=bgy)){
        IgnoreDoubleClick=true;
        if(num==Min) CurY-=bgh/4;
        TagY-=bgh;
      }
      if(((bgy+bgh)<y)&&(y<=(bgy+(bgh*2)))){
        IgnoreDoubleClick=true;
        if(num==Max) CurY+=bgh/4;
        TagY+=bgh;
      }
    }
  }
  
  s32 num=(TagY+(bgh/2))/bgh;
  bool sf=false;
  if(num<Min){
    num=Min;
    sf=true;
  }
  if(Max<num){
    num=Max;
    sf=true;
  }
  if(sf==true) SoundSE_Start(WAVFN_AutoFlag);
  
  TagY=num*bgh;
  
  StoreMouseLog(MouseLogs[2].mx-x,MouseLogs[2].my-y,mv);
  mv=0;
  
//  _consolePrintf("%d,%d %d,%d %d,%d,%d\n",MouseLogs[0].mx,MouseLogs[0].my,MouseLogs[2].mx,MouseLogs[2].my,MouseLogs[0].mv,MouseLogs[1].mv,MouseLogs[2].mv);
  
  if(IgnoreDoubleClick==false){
    if((MouseLogs[0].mx<16)&&(MouseLogs[0].my<16)&&(MouseLogs[2].mx<16)&&(MouseLogs[2].my<16)){
      if((MouseLogs[0].mv<15)&&(MouseLogs[1].mv<30)&&(MouseLogs[2].mv<15)){
        Disabled();
        return;
      }
    }
  }
}

void CNumBox::VsyncUpdate(u32 VsyncCount)
{
  mv+=VsyncCount;
  
  const s32 bgh=FontHeight;
  
  s32 numold=(CurY+(bgh/2))/bgh;
  
  float tagy=TagY;
  tagy+=0.5;
  float v=(tagy-CurY)/8;
  
  if((-0.05<v)&&(v<0.05)){
    CurY=tagy;
    }else{
    CurY+=v;
  }
  
  s32 numnew=(CurY+(bgh/2))/bgh;
  
  if((Min<=numnew)&&(numnew<=Max)){
    if((Min<=numold)&&(numold<=Max)){
      if(numold!=numnew){
        if(v<0) v=-v;
        v=v*(v/2)*(v/2);
        s32 vol=128-(v/6);
        if(vol<0) vol=0;
        if(127<vol) vol=127;
        SoundSE_StartVol(WAVFN_AutoOpen,vol);
      }
    }
  }
}

bool CNumBox::GetEnabled(void)
{
  return(Enabled);
}

void CNumBox::Draw(CglCanvas *pcan)
{ 
  if(Enabled==false){
    pClosedBG->BitBlt(pcan,ClosedBGRect.x,ClosedBGRect.y);
    
    const s32 Num=CurY/FontHeight;
    const s32 x=PosLeft,y=PosTop;
    const s32 w=FontWidth;
    if(Num<10){
      pFont->BitBltLimitY(pcan,x-(w/2),y,FontHeight,Num*FontHeight);
      }else{
      if(Num<100){
        pFont->BitBltLimitY(pcan,x-w,y,FontHeight,(Num/10)*FontHeight);
        pFont->BitBltLimitY(pcan,x,y,FontHeight,(Num%10)*FontHeight);
        }else{
        pFont->BitBltLimitY(pcan,x-w-(w/2),y,FontHeight,(Num/100)*FontHeight);
        pFont->BitBltLimitY(pcan,x-(w/2),y,FontHeight,(Num/10%10)*FontHeight);
        pFont->BitBltLimitY(pcan,x+(w/2),y,FontHeight,(Num%10)*FontHeight);
      }
    }
    return;
  }
  
  {
    const TRect r=OpenBGRect;
    pWindowBG->BitBlt(pcan,r.x,r.y,r.w,r.h,0,0,false);
  }
  
  for(s32 idx=-FontHeight;idx<FontHeight*2;idx++){
    s32 SrcY=CurY+idx;
    
    u32 Level32=0;
    if(idx<0){
      Level32=32+idx;
      }else{
      if(idx<32){
        Level32=32;
        }else{
        Level32=32-(idx-32);
      }
    }
    
    if(3<=Level32){
      if(Level32!=32) Level32-=2;
      const s32 Num=SrcY/FontHeight;
      const s32 NumOfs=SrcY%FontHeight;
      if((Min<=Num)&&(Num<=Max)){
        const s32 x=PosLeft,y=PosTop;
        const s32 w=FontWidth;
        if(Num<10){
          pFont->BitBlt1Line_SetLevel(pcan,x-(w/2),y+idx,(Num*FontHeight)+NumOfs,Level32);
          }else{
          if(Num<100){
            pFont->BitBlt1Line_SetLevel(pcan,x-w,y+idx,((Num/10)*FontHeight)+NumOfs,Level32);
            pFont->BitBlt1Line_SetLevel(pcan,x,y+idx,((Num%10)*FontHeight)+NumOfs,Level32);
            }else{
            pFont->BitBlt1Line_SetLevel(pcan,x-w-(w/2),y+idx,((Num/100)*FontHeight)+NumOfs,Level32);
            pFont->BitBlt1Line_SetLevel(pcan,x-(w/2),y+idx,((Num/10%10)*FontHeight)+NumOfs,Level32);
            pFont->BitBlt1Line_SetLevel(pcan,x+(w/2),y+idx,((Num%10)*FontHeight)+NumOfs,Level32);
          }
        }
      }
    }
  }
}

