
#include <nds.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "_console.h"
#include "_consolewritelog.h"
#include "maindef.h"
#include "memtool.h"
#include "_const.h"
#include "lang.h"
#include "datetime.h"
#include "inifile.h"

#include "glib.h"
#include "skin.h"
#include "strtool.h"

#include "fat2.h"
#include "shell.h"
#include "sndeff.h"
#include "procstate.h"
#include "strpcm.h"
#include "skinpack.h"

#include "cfont.h"

#include "rect.h"
#include "libms.h"
#include "Particle.h"

static int rand(int max)
{
  int ret=rand()/((double)RAND_MAX+1)*max;
  return(ret);
}

static CglTGF *pTitleTop;

static CglB15 *pBG,*pTile;
static CglB15 *pBGMasked,*pTileMasked;

static u32 ScrollIndex;
static const u32 ScrollIndexDiv=4;
static const u32 TileSize=128;

enum EPage {EP_Select,EP_Custom};
static EPage Page;

// --------------------------------------------------------------------

// ------------------------------------------------------------------

#include "proc_Title_HighScore.h"

// ------------------------------------------------------------------

static bool MSHelp_Inited;

static void MSHelp_Free(void)
{
  if(MSHelp_Inited==false) return;
  MSHelp_Inited=false;
  MS_Free();
}

static void MSHelp_Init(void)
{
  MSHelp_Free();
  
  switch(Page){
    case EP_Select: {
      const u32 size=20;
      MS_Init(0,size,size*3/4,20,false,0);
    } break;
    case EP_Custom: {
      MS_Init(0,ProcState.Main.CustomWidth+2,ProcState.Main.CustomHeight+2,ProcState.Main.CustomBombs,false,0);
      MS_SetupDemo();
    } break;
  }
  
  MS_SetEnableSE(false);
  
  MSHelp_Inited=true;
}

static void MSHelp_Vsync(void)
{
  switch(Page){
    case EP_Select: {
      while(1){
        static u32 waitcnt=20;
        if(MS_ProcessAutoOpen()==true){
          waitcnt++;
          }else{
          if(rand(2)==0){
            if(waitcnt!=0){
              waitcnt--;
              return;
              }else{
              if(MS_TitleDemo()==false){
                if(rand(200)==0) MSHelp_Init();
                return;
              }
            }
            }else{
            return;
          }
        }
      }
    } break;
    case EP_Custom: {
    } break;
  }
}

// ------------------------------------------------------------------

static void DrawTopScreen(void);
static void DrawBottomScreen(void);
static void DrawSelectedBottomScreen(u32 SelectedLevel);

#include "proc_Title_Buttons.h"

#include "libnumbox.h"

#include "proc_Title_Page_Select.h"
#include "proc_Title_Page_Custom.h"

// ------------------------------------------------------------------

static void DrawTopScreen(void)
{
  if(HighScore_ReqDraw()==true){
    CglCanvas *pcan=new CglCanvas(&MM_Temp,NULL,ScreenWidth,ScreenHeight,pf15bit);
    HighScore_Draw(pcan);
    pcan->BitBltFullBeta(pScreenSub->pCanvas);
    if(pcan!=NULL){
      delete pcan; pcan=NULL;
    }
    return;
  }
  
  if(MS_Refresh()==false) return;
  
  CglCanvas *pcan=new CglCanvas(&MM_Temp,NULL,ScreenWidth,ScreenHeight,pf15bit);
  pcan->SetCglFont(pCglFontDefault);
  
  MS_Draw_Top(pcan,false);
  
  switch(Page){
    case EP_Select: {
      Select_DrawTopScreen(pcan);
    } break;
    case EP_Custom: {
      Custom_DrawTopScreen(pcan);
    } break;
  }
  
  {
    pcan->SetFontTextColor(RGB15(16,16,16)|BIT15);
    u32 x=ScreenWidth-2,y=2;
    char msg[256];
    snprintf(msg,256,"%s %s",ROMTITLE,ROMVERSION);
    pcan->TextOutA(x-pcan->GetTextWidthA(msg),y,msg);
  }
  
  pcan->BitBltFullBeta(pScreenSub->pCanvas);
  
  if(pcan!=NULL){
    delete pcan; pcan=NULL;
  }
}

static void DrawBottomScreen(void)
{
  CglCanvas *pcan=pScreenMain->pBackCanvas;
  
  bool AddBG=true;
  if(Page==EP_Custom) AddBG=false;
  
  if(AddBG==false){
    pBG->pCanvas->BitBltFullBeta(pcan);
    }else{
    const u32 ofs=(ScrollIndex/ScrollIndexDiv)&~1;
    const u32 xofs=ofs/2,yofs=(TileSize-1)-ofs;
    u32 *pbuf=(u32*)pcan->GetVRAMBuf();
    const u32 *psrc=(u32*)pBGMasked->pCanvas->GetVRAMBuf();
    for(u32 y=0;y<ScreenHeight;y++){
      const u32 *padd=(u32*)pTileMasked->pCanvas->GetScanLine((yofs+y)&(TileSize-1));
      for(u32 x=0;x<ScreenWidth/2;x++){
        u32 col=(*psrc++)+padd[(xofs+x)&((TileSize/2)-1)];
        *pbuf++=col|BIT15|BIT31;
      }
    }
  }
  
  Particle_Draw(pcan);
  
  switch(Page){
    case EP_Select: {
      Select_DrawBottomScreen(pcan);
    } break;
    case EP_Custom: {
      Custom_DrawBottomScreen(pcan);
    } break;
  }
  
  Btns_Draw(pcan);
}

static void DrawSelectedBottomScreen(u32 SelectedLevel)
{
  CglCanvas *pcan=pScreenMain->pBackCanvas;
  
  pBG->pCanvas->BitBltFullBeta(pcan);
  
  switch(Page){
    case EP_Select: {
      Select_DrawSelectedBottomScreen(SelectedLevel,pcan);
    } break;
    case EP_Custom: {
      Custom_DrawSelectedBottomScreen(SelectedLevel,pcan);
    } break;
  }
}

static void CB_KeyPress(u32 VsyncCount,u32 Keys,bool FirstFlag)
{
  switch(Page){
    case EP_Select: {
      Select_CB_KeyPress(VsyncCount,Keys,FirstFlag);
    } break;
    case EP_Custom: {
      Custom_CB_KeyPress(VsyncCount,Keys,FirstFlag);
    } break;
  }
}

static bool MouseDownFlag;
static EBtnFunc MouseSelBtnFunc;
static s32 mx,my;

static void CB_MouseDown(s32 x,s32 y)
{
  if(Page==EP_Custom){
    if(Custom_CB_MouseDown(x,y)==true) return;
  }
  
  MouseDownFlag=true;
  
  for(s32 idx=0;idx<8;idx++){
    Particle_SetCursor(x,y,0,0);
  }
  
  MouseSelBtnFunc=Btns_GetFunc(x,y);
  
  mx=x;
  my=y;
}

static void CB_MouseMove(s32 x,s32 y)
{
  if(Page==EP_Custom){
    if(Custom_CB_MouseMove(x,y)==true) return;
  }
  
  if(MouseDownFlag==false) return;
  
  const s32 vx=mx-x,vy=my-y;
  
  const s32 cnt=sqrt((float)((vx*vx)+(vy*vy)));
  for(s32 idx=0;idx<cnt;idx++){
    u32 px=mx-(vx*idx/cnt),py=my-(vy*idx/cnt);
    Particle_SetCursor(px,py,vx,vy);
  }
  
  mx=x;
  my=y;
}

static void CB_MouseUp(s32 x,s32 y)
{
  if(Page==EP_Custom){
    if(Custom_CB_MouseUp(x,y)==true) return;
  }
  
  if(MouseDownFlag==false) return;
  MouseDownFlag=false;
  
  if(MouseSelBtnFunc!=Btns_GetFunc(x,y)) return;
  
  Btns_Execute(MouseSelBtnFunc);
}

static void CB_Start(void)
{
  MouseDownFlag=false;
  
  SoundBGM_Start(WAVFN_BGMTitle,false);
  
  Particle_Init(false);
  
  HighScore_Init();
  if(ProcState.Main.TitleShowHighScore==true) HighScore_Fixed();
  
  pTitleTop=SkinPack_LoadTGF(ESF_TitleTop_tgf);
  
  pBG=SkinPack_LoadB15(ESF_LevelSelect_BG_b15);
  pTile=SkinPack_LoadB15(ESF_LevelSelect_Tile_b15);
  
  pBGMasked=SkinPack_LoadB15(ESF_LevelSelect_BG_b15);
  {
    u32 *pbuf=(u32*)pBGMasked->pCanvas->GetVRAMBuf();
    const u32 cnt=pBGMasked->GetWidth()*pBGMasked->GetHeight()/2;
    for(u32 idx=0;idx<cnt;idx++){
      *pbuf++=*pbuf&0x7fff7fff;
    }
  }
  
  pTileMasked=SkinPack_LoadB15(ESF_LevelSelect_Tile_b15);
  {
    u32 *pbuf=(u32*)pTileMasked->pCanvas->GetVRAMBuf();
    const u32 cnt=pTileMasked->GetWidth()*pTileMasked->GetHeight()/2;
    for(u32 idx=0;idx<cnt;idx++){
      *pbuf++=*pbuf&0x7fff7fff;
    }
  }
  
  if(ProcState.Main.PlayLevel==0){
    Page=EP_Custom;
    }else{
    Page=EP_Select;
  }
  
  MSHelp_Inited=false;
  MSHelp_Init();
  
  Btns_Init();
  Btns_Setup();
  
  Select_Init();
  Custom_Init();
  
  DrawTopScreen();
  
  ScrollIndex=0;
  DrawBottomScreen();
  
  ScreenMain_Flip_ProcFadeEffect();
}

static void CB_VsyncUpdate(u32 VsyncCount)
{
  if(IPC6->PanelOpened==false) return;
  
  if(HighScore_VsyncUpdate(VsyncCount)==false){
    DrawTopScreen();
    return;
    }else{
    static bool a=true;
    if(HighScore_ReqDraw()==false){
      a=true;
      }else{
      if(a==true){
        a=false;
        DrawTopScreen();
        MS_ExclusiveRefresh();
      }
    }
  }
  
  if(HighScore_ReqDraw()==false){
    MSHelp_Vsync();
    DrawTopScreen();
  }
  
  if(Custom_CB_VsyncUpdate(VsyncCount)==true) return;
  
  for(u32 idx=0;idx<VsyncCount;idx++){
    Particle_Process();
  }
  
  ScrollIndex+=VsyncCount;
  ScrollIndex&=(TileSize*ScrollIndexDiv)-1;
  
  DrawBottomScreen();
  
  ScreenMain_Flip_ProcFadeEffect();
}

static void CB_End(void)
{
  if(pTitleTop!=NULL){
    delete pTitleTop; pTitleTop=NULL;
  }
  
  if(pBG!=NULL){
    delete pBG; pBG=NULL;
  }
  if(pTile!=NULL){
    delete pTile; pTile=NULL;
  }
  if(pBGMasked!=NULL){
    delete pBGMasked; pBGMasked=NULL;
  }
  if(pTileMasked!=NULL){
    delete pTileMasked; pTileMasked=NULL;
  }
  
  Select_Free();
  Custom_Free();
  
  HighScore_Free();
  
  ProcState_Save();
  
  Particle_Free();
  
  SoundBGM_Stop();
  
  Btns_Free();
  
  MSHelp_Free();
}

static void CB_ReqSoftReset(void)
{
  pScreenSub->pCanvas->FillFull(0);
  
  pScreenMain->pBackCanvas->FillFull(RGB15(0,0,0)|BIT15);
  SetProcFadeEffect(EPFE_UpToDown);
  ScreenMain_Flip_ProcFadeEffect();
  pScreenMain->pBackCanvas->FillFull(RGB15(0,0,0)|BIT15);
  
  extern bool ReqSoftResetFlag;
  ReqSoftResetFlag=true;
}

static void CB_PanelOpen(void)
{
//  SoundSE_Start(WAVFN_PanelOpen);
}

static void CB_PanelClose(void)
{
  SoundSE_Start(WAVFN_PanelClose);
}

void ProcTitle_SetCallBack(TCallBack *pCallBack)
{
  pCallBack->Start=CB_Start;
  pCallBack->VsyncUpdate=CB_VsyncUpdate;
  pCallBack->End=CB_End;
  pCallBack->KeyPress=CB_KeyPress;
  pCallBack->MouseDown=CB_MouseDown;
  pCallBack->MouseMove=CB_MouseMove;
  pCallBack->MouseUp=CB_MouseUp;
  pCallBack->ReqSoftReset=CB_ReqSoftReset;
  pCallBack->PanelOpen=CB_PanelOpen;
  pCallBack->PanelClose=CB_PanelClose;
}

