
#include <math.h>

static u32 HighScore_LastPlayLevel;

static CglB15 *pHighScoreBG;

static CglTGF *pHighScoreFont16TGF,*pHighScoreFont24TGF;
static CFont *pHighScoreFont16,*pHighScoreFont24;

static u32 HighScore_TitlePos;

#define HighScore_MaxLength (21)
#define HighScore_MaxLine (5)

typedef struct {
  char ch;
  float x,y;
  float tx,ty;
} THSItem;
static u32 HSItemsCount;
#define HSItemsCountMax (HighScore_MaxLength*HighScore_MaxLine)
static THSItem HSItems[HSItemsCountMax];

static void HishScore_Setup_ins_AddLine(float x,float y,const char *pstr)
{
  float OfsX=x;
  float MulX=ScreenWidth/(ScreenWidth-(OfsX*2));
  
  while(1){
    char ch=*pstr++;
    if(ch==0) break;
    
    if(ch!=' '){
      THSItem *phsi=&HSItems[HSItemsCount++];
      phsi->ch=ch;
      phsi->x=(1-cos((x-OfsX)*MulX/ScreenWidth*3.14159))/2*ScreenWidth;
      phsi->y=ScreenHeight-24;
      phsi->tx=x;
      phsi->ty=y;
    }
    
    x+=pHighScoreFont16->GetCharWidth(ch);
  }
}

static void HighScore_Setup(void)
{
  HighScore_TitlePos=0;
  
  HSItemsCount=0;
  
  const u32 x=(ScreenWidth-pHighScoreFont16->GetTextWidth("123456789012345678901"))/2;
  u32 y=8;
  
  y+=24;
  y+=16;
  
  char *plvmsg[6]={"Cust","Prac","Easy","Norm","Hard",""};
  
  const TProcState_Score *pScore=&ProcState.Score;
  
  u32 cnt=HighScore_MaxLine;
  if(pScore->ClearCount[4]==0){
    cnt--;
    y+=10;
  }
  
  for(u32 idx=0;idx<cnt;idx++){
    u32 lv=idx+1;
    char lvc='0'+lv;
    if(lv==5) lvc='X';
    u32 Play=pScore->PlayCount[lv],Clear=pScore->ClearCount[lv],Sec=pScore->FastTime[lv]/60,RestBomb=pScore->RestBomb[lv];
    if(999<Play) Play=999;
    if(999<Clear) Clear=999;
    if(99999<Sec) Sec=99999;
    if(999<RestBomb) RestBomb=999;
    char msg[32];
    if(Play==0){
      snprintf(msg,32,"Lv%c %-4s %-12s",lvc,plvmsg[lv],"  NO DATA");
      }else{
      if(Clear==0){
        snprintf(msg,32,"Lv%c %-4s Rest%-3d /%3d",lvc,plvmsg[lv],RestBomb,Play);
        }else{
        snprintf(msg,32,"Lv%c %-4s%5ds%3d/%3d",lvc,plvmsg[lv],Sec,Clear,Play);
      }
    }
    HishScore_Setup_ins_AddLine(x,y,msg);
    y+=28;
  }
}

static void HighScore_Init(void)
{
  HighScore_LastPlayLevel=ProcState.Main.PlayLevel;
  
  pHighScoreBG=SkinPack_LoadB15(ESF_HighScoreBG_b15);
  
  pHighScoreFont16TGF=SkinPack_LoadTGF(ESF_HighScoreFont16_tgf);
  pHighScoreFont24TGF=SkinPack_LoadTGF(ESF_HighScoreFont24_tgf);
  pHighScoreFont16=new CFont(CFont::EBM_TGF,NULL,pHighScoreFont16TGF,DefaultImplantCharsCount,DefaultImplantChars);
  pHighScoreFont24=new CFont(CFont::EBM_TGF,NULL,pHighScoreFont24TGF,DefaultImplantCharsCount,DefaultImplantChars);
  
  HighScore_Setup();
}

static void HighScore_Fixed(void)
{
  HighScore_TitlePos=8;
  
  for(u32 idx=0;idx<HSItemsCount;idx++){
    THSItem *phsi=&HSItems[idx];
    phsi->x=phsi->tx;
    phsi->y=phsi->ty;
  }
}

static void HighScore_Free(void)
{
  if(pHighScoreFont16!=NULL){
    delete pHighScoreFont16; pHighScoreFont16=NULL;
  }
  if(pHighScoreFont24!=NULL){
    delete pHighScoreFont24; pHighScoreFont24=NULL;
  }
  if(pHighScoreFont16TGF!=NULL){
    delete pHighScoreFont16TGF; pHighScoreFont16TGF=NULL;
  }
  if(pHighScoreFont24TGF!=NULL){
    delete pHighScoreFont24TGF; pHighScoreFont24TGF=NULL;
  }
  
  if(pHighScoreBG!=NULL){
    delete pHighScoreBG; pHighScoreBG=NULL;
  }
}

static bool HighScore_VsyncUpdate(u32 VsyncCount)
{
  if(ProcState.Main.TitleShowHighScore==false) return(true);
  
  bool endf=true;
  
  for(u32 idx=0;idx<VsyncCount;idx++){
    if(HighScore_TitlePos!=8) HighScore_TitlePos++;
    for(u32 idx=0;idx<HSItemsCount;idx++){
      THSItem *phsi=&HSItems[idx];
      float mx=(phsi->tx-phsi->x)/8;
      float my=(phsi->ty-phsi->y)/8;
      float Lim=0.1;
      if((mx<-Lim)||(Lim<mx)||(my<-Lim)||(Lim<my)) endf=false;
      float v=1-((phsi->tx/ScreenWidth)/2);
      phsi->x+=mx*v;
      phsi->y+=my*v;
    }
  }
  
  return(endf);
}

static bool HighScore_ReqDraw(void)
{
  if(ProcState.Main.TitleShowHighScore==false) return(false);
  if(Page!=EP_Select) return(false);
  return(true);
}

static void HighScore_Draw(CglCanvas *pcan)
{
  pHighScoreBG->pCanvas->BitBltFullBeta(pcan);
  
  {
    char msg[32];
    snprintf(msg,32,"- HIGH SCORE -");
    pHighScoreFont24->DrawText(pcan,(ScreenWidth-pHighScoreFont24->GetTextWidth(msg))/2,HighScore_TitlePos,msg);
  }
  
  for(u32 idx=0;idx<HSItemsCount;idx++){
    THSItem *phsi=&HSItems[idx];
    pHighScoreFont16->DrawChar(pcan,phsi->x,phsi->y,phsi->ch);
  }
}


