
static const u32 NumBoxsCount=3;
static CNumBox *pNumBoxs[NumBoxsCount];
static CNumBox *pNumBox_ActivePointer;
static u32 NumBox_ActiveIndex;

static CglTGF *pCustom_OverlayBG;

typedef struct {
  u32 WidthMin,WidthMax;
  u32 HeightMin,HeightMax;
  u32 BombsMin,BombsMax;
} TCustomLimit;
static TCustomLimit CustomLimit;

static void CustomLimit_Refresh(void)
{
  TCustomLimit *pcl=&CustomLimit;
  
  pcl->WidthMin=MS_CellsMinWidth-2;
  pcl->WidthMax=MS_CellsMaxWidth-2;
  pcl->HeightMin=MS_CellsMinHeight-2;
  pcl->HeightMax=MS_CellsMaxHeight-2;
  
  u32 Width=ProcState.Main.CustomWidth,Height=ProcState.Main.CustomHeight;
  
  pcl->BombsMin=2;
  pcl->BombsMax=(Width*Height)-(3*3);
}

static void Custom_Init(void)
{
  pCustom_OverlayBG=SkinPack_LoadTGF(ESF_Custom_OverlayBG_tgf);
  
  NumBox_Init();
  pNumBoxs[0]=new CNumBox(2,48,50,ProcState.Main.CustomWidth);
  pNumBoxs[1]=new CNumBox(2,144,50,ProcState.Main.CustomHeight);
  pNumBoxs[2]=new CNumBox(3,116,110,ProcState.Main.CustomBombs);
  pNumBox_ActivePointer=NULL;
  NumBox_ActiveIndex=(u32)-1;
}

static void Custom_Free(void)
{
  if(pCustom_OverlayBG!=NULL){
    delete pCustom_OverlayBG; pCustom_OverlayBG=NULL;
  }
  
  for(u32 idx=0;idx<NumBoxsCount;idx++){
    if(pNumBoxs[idx]!=NULL){
      delete pNumBoxs[idx]; pNumBoxs[idx]=NULL;
    }
  }
  pNumBox_ActivePointer=NULL;
  NumBox_ActiveIndex=(u32)-1;
  NumBox_Free();
}

static void Custom_CB_KeyPress(u32 VsyncCount,u32 Keys,bool FirstFlag)
{
  if(Keys&(KEY_START|KEY_A)){
    Btns_Execute(EBF_CustomStart);
  }
  if(Keys&(KEY_SELECT|KEY_B)){
    Btns_Execute(EBF_CustomBack);
  }
}

static bool Custom_CB_MouseDown(s32 x,s32 y)
{
  if(pNumBox_ActivePointer!=NULL){
    pNumBox_ActivePointer->CB_MouseDown(x,y);
    return(true);
  }
  
  CustomLimit_Refresh();
  for(u32 idx=0;idx<NumBoxsCount;idx++){
    u32 Min=0,Max=0;
    switch(idx){
      case 0: Min=CustomLimit.WidthMin; Max=CustomLimit.WidthMax; break;
      case 1: Min=CustomLimit.HeightMin; Max=CustomLimit.HeightMax; break;
      case 2: Min=CustomLimit.BombsMin; Max=CustomLimit.BombsMax; break;
    }
    pNumBoxs[idx]->SetLimit(Min,Max);
    if(pNumBoxs[idx]->CB_MouseDown(x,y)==true){
      _consolePrintf("Open NumBox %d: %d,%d.\n",idx,Min,Max);
      pNumBox_ActivePointer=pNumBoxs[idx];
      NumBox_ActiveIndex=idx;
      return(true);
    }
  }
  
  return(false);
}

static bool Custom_CB_MouseMove(s32 x,s32 y)
{
  if(pNumBox_ActivePointer!=NULL){
    pNumBox_ActivePointer->CB_MouseMove(x,y);
    return(true);
  }
  
  return(false);
}

static bool Custom_CB_MouseUp(s32 x,s32 y)
{
  if(pNumBox_ActivePointer!=NULL){
    pNumBox_ActivePointer->CB_MouseUp(x,y);
    if(pNumBox_ActivePointer->GetEnabled()==false){
      u32 num=pNumBox_ActivePointer->GetCurrentNumber();
      float BombsPer=(float)ProcState.Main.CustomBombs/(ProcState.Main.CustomWidth*ProcState.Main.CustomHeight);
      _consolePrintf("Closed NumBox %d: %d.\n",NumBox_ActiveIndex,num);
      _consolePrintf("BombsPer: %f.\n",BombsPer);
      bool AjustBombs=false;
      switch(NumBox_ActiveIndex){
        case 0: ProcState.Main.CustomWidth=num; AjustBombs=true; break;
        case 1: ProcState.Main.CustomHeight=num; AjustBombs=true; break;
        case 2: ProcState.Main.CustomBombs=num; break;
      }
      if(AjustBombs==true){
        ProcState.Main.CustomBombs=(ProcState.Main.CustomWidth*ProcState.Main.CustomHeight)*BombsPer;
      }
      CustomLimit_Refresh();
      if(ProcState.Main.CustomBombs<CustomLimit.BombsMin) ProcState.Main.CustomBombs=CustomLimit.BombsMin;
      if(CustomLimit.BombsMax<ProcState.Main.CustomBombs) ProcState.Main.CustomBombs=CustomLimit.BombsMax;
      pNumBoxs[2]->SetCurrentNumber(ProcState.Main.CustomBombs);
      ProcState_RequestSave=true;
      MSHelp_Init();
      DrawTopScreen();
      DrawBottomScreen();
      pNumBox_ActivePointer=NULL;
      NumBox_ActiveIndex=(u32)-1;
    }
    return(true);
  }
  
  return(false);
}

static void Custom_DrawTopScreen(CglCanvas *pcan)
{
}

static void Custom_DrawBottomScreen(CglCanvas *pcan)
{
  pCustom_OverlayBG->BitBlt(pcan,0,0);
  
  for(u32 idx=0;idx<NumBoxsCount;idx++){
    pNumBoxs[idx]->Draw(pScreenMain->pBackCanvas);
  }
  
  const u32 CellsCount=ProcState.Main.CustomWidth*ProcState.Main.CustomHeight;
  const u32 Per=((ProcState.Main.CustomBombs*100)+(CellsCount/2))/CellsCount;
  
  CglTGF *pFont=NumBox_GetFont();
  const u32 FontWidth=pFont->GetWidth(),FontHeight=pFont->GetHeight()/10;
  
  u32 x=176,y=110;
  if(10<=Per){
    pFont->BitBltLimitY(pcan,x,y,FontHeight,(Per/10)*FontHeight);
  }
  x+=FontWidth;
  pFont->BitBltLimitY(pcan,x,y,FontHeight,(Per%10)*FontHeight);
  
}

static void Custom_DrawSelectedBottomScreen(u32 SelectedLevel,CglCanvas *pcan)
{
}

static bool Custom_CB_VsyncUpdate(u32 VsyncCount)
{
  for(u32 idx=0;idx<NumBoxsCount;idx++){
    pNumBoxs[idx]->VsyncUpdate(VsyncCount);
  }
  
  if(pNumBox_ActivePointer!=NULL){
    pNumBox_ActivePointer->Draw(pScreenMain->pViewCanvas);
    return(true);
  }
  
  return(false);
}

